@@ -1,14 +1,14 @@
-#!/usr/bin/env python
-# -*- coding: utf-8 -*-
-from __future__ import unicode_literals
+import io
 import json
 import errno
-from datetime import datetime
 
-from mock import Mock, patch, call
+from freezegun import freeze_time
 from sure import expect
 
 from httpretty.core import HTTPrettyRequest, FakeSSLSocket, fakesock, httpretty
+from httpretty.core import URIMatcher, URIInfo
+
+from tests.compat import Mock, patch, call
 
 
 class SocketErrorStub(Exception):
@@ -23,11 +23,11 @@ def test_request_stubs_internals():
     # Given a valid HTTP request header string
     headers = "\r\n".join([
         'POST /somewhere/?name=foo&age=bar HTTP/1.1',
-        'Accept-Encoding: identity',
-        'Host: github.com',
-        'Content-Type: application/json',
-        'Connection: close',
-        'User-Agent: Python-urllib/2.7',
+        'accept-encoding: identity',
+        'host: github.com',
+        'content-type: application/json',
+        'connection: close',
+        'user-agent: Python-urllib/2.7',
     ])
 
     # When I create a HTTPrettyRequest with an empty body
@@ -42,17 +42,18 @@ def test_request_stubs_internals():
         'user-agent': 'Python-urllib/2.7'
     })
 
-    # And the `rfile` should be a StringIO
-    request.should.have.property('rfile').being.a('StringIO.StringIO')
+    # And the `rfile` should be a io.BytesIO
+    type_as_str = io.BytesIO.__module__ + '.' + io.BytesIO.__name__
 
-    # And the `wfile` should be a StringIO
-    request.should.have.property('wfile').being.a('StringIO.StringIO')
+    request.should.have.property('rfile').being.a(type_as_str)
+
+    # And the `wfile` should be a io.BytesIO
+    request.should.have.property('wfile').being.a(type_as_str)
 
     # And the `method` should be available
     request.should.have.property('method').being.equal('POST')
 
 
-
 def test_request_parse_querystring():
     ("HTTPrettyRequest#parse_querystring should parse unicode data")
 
@@ -151,22 +152,23 @@ def test_request_parse_body_when_unrecognized():
 
 
 def test_request_string_representation():
-    ("HTTPrettyRequest should have a debug-friendly "
+    ("HTTPrettyRequest should have a forward_and_trace-friendly "
      "string representation")
 
     # Given a request string containing a unicode encoded querystring
     headers = "\r\n".join([
         'POST /create HTTP/1.1',
         'Content-Type: JPEG-baby',
+        'Host: blog.falcao.it'
     ])
     # And a valid urlencoded body
     body = "foobar:\nlalala"
 
     # When I create a HTTPrettyRequest with that data
-    request = HTTPrettyRequest(headers, body)
+    request = HTTPrettyRequest(headers, body, sock=Mock(is_https=True))
 
     # Then its string representation should show the headers and the body
-    str(request).should.equal('<HTTPrettyRequest("JPEG-baby", total_headers=1, body_length=14)>')
+    str(request).should.equal('<HTTPrettyRequest("POST", "https://blog.falcao.it/create", headers={\'Content-Type\': \'JPEG-baby\', \'Host\': \'blog.falcao.it\'}, body=14)>')
 
 
 def test_fake_ssl_socket_proxies_its_ow_socket():
@@ -186,17 +188,14 @@ def test_fake_ssl_socket_proxies_its_ow_socket():
     socket.send.assert_called_once_with("FOO")
 
 
-@patch('httpretty.core.datetime')
-def test_fakesock_socket_getpeercert(dt):
+@freeze_time("2013-10-04 04:20:00")
+def test_fakesock_socket_getpeercert():
     ("fakesock.socket#getpeercert should return a hardcoded fake certificate")
-    # Background:
-    dt.now.return_value = datetime(2013, 10, 4, 4, 20, 0)
-
     # Given a fake socket instance
     socket = fakesock.socket()
 
-    # And that it's bound to some host and port
-    socket.connect(('somewhere.com', 80))
+    # And that it's bound to some host
+    socket._host = 'somewhere.com'
 
     # When I retrieve the peer certificate
     certificate = socket.getpeercert()
@@ -209,7 +208,7 @@ def test_fakesock_socket_getpeercert(dt):
             ((u'organizationalUnitName', u'Domain Control Validated'),),
             ((u'commonName', u'*.somewhere.com'),)),
         u'subjectAltName': (
-            (u'DNS', u'*somewhere.com'),
+            (u'DNS', u'*.somewhere.com'),
             (u'DNS', u'somewhere.com'),
             (u'DNS', u'*')
         )
@@ -231,7 +230,6 @@ def test_fakesock_socket_ssl():
     result.should.equal(sentinel)
 
 
-
 @patch('httpretty.core.old_socket')
 @patch('httpretty.core.POTENTIAL_HTTP_PORTS')
 def test_fakesock_socket_connect_fallback(POTENTIAL_HTTP_PORTS, old_socket):
@@ -249,26 +247,21 @@ def test_fakesock_socket_connect_fallback(POTENTIAL_HTTP_PORTS, old_socket):
     # Then it should have open a real connection in the background
     old_socket.return_value.connect.assert_called_once_with(('somewhere.com', 42))
 
-    # And _closed is set to False
-    socket._closed.should.be.false
-
 
 @patch('httpretty.core.old_socket')
 def test_fakesock_socket_close(old_socket):
     ("fakesock.socket#close should close the actual socket in case "
-     "it's not http and _closed is False")
+     "it's not http and __truesock_is_connected__ is True")
     # Given a fake socket instance that is synthetically open
     socket = fakesock.socket()
-    socket._closed = False
+    socket.__truesock_is_connected__ = True
 
     # When I close it
     socket.close()
 
     # Then its real socket should have been closed
     old_socket.return_value.close.assert_called_once_with()
-
-    # And _closed is set to True
-    socket._closed.should.be.true
+    socket.__truesock_is_connected__.should.be.false
 
 
 @patch('httpretty.core.old_socket')
@@ -296,117 +289,153 @@ def test_fakesock_socket_makefile(old_socket):
 
 @patch('httpretty.core.old_socket')
 def test_fakesock_socket_real_sendall(old_socket):
+    ("fakesock.socket#real_sendall calls truesock#connect and bails "
+     "out when not http")
+    # Background: the real socket will stop returning bytes after the
+    # first call
+    real_socket = old_socket.return_value
+    real_socket.recv.side_effect = [b'response from server', b""]
+
+    # Given a fake socket
+    socket = fakesock.socket()
+    socket._address = ('1.2.3.4', 42)
+
+    # When I call real_sendall with data, some args and kwargs
+    socket.real_sendall(b"SOMEDATA", b'some extra args...', foo=b'bar')
+
+    # Then it should have called sendall in the real socket
+    real_socket.sendall.assert_called_once_with(b"SOMEDATA", b'some extra args...', foo=b'bar')
+
+    # # And setblocking was never called
+    # real_socket.setblocking.called.should.be.false
+
+    # And recv was never called
+    real_socket.recv.called.should.be.false
+
+    # And the buffer is empty
+    socket.fd.read().should.equal(b'')
+
+    # And connect was never called
+    real_socket.connect.called.should.be.false
+
+
+@patch('httpretty.core.old_socket')
+def test_fakesock_socket_real_sendall_when_http(old_socket):
     ("fakesock.socket#real_sendall sends data and buffers "
      "the response in the file descriptor")
     # Background: the real socket will stop returning bytes after the
     # first call
     real_socket = old_socket.return_value
-    real_socket.recv.side_effect = ['response from server', ""]
+    real_socket.recv.side_effect = [b'response from server', b""]
 
     # Given a fake socket
     socket = fakesock.socket()
+    socket._address = ('1.2.3.4', 42)
+    socket.is_http = True
 
     # When I call real_sendall with data, some args and kwargs
-    socket.real_sendall("SOMEDATA", 'some extra args...', foo='bar')
+    socket.real_sendall(b"SOMEDATA", b'some extra args...', foo=b'bar')
 
     # Then it should have called sendall in the real socket
-    real_socket.sendall.assert_called_once_with("SOMEDATA", 'some extra args...', foo='bar')
+    real_socket.sendall.assert_called_once_with(b"SOMEDATA", b'some extra args...', foo=b'bar')
 
-    # And the timeout was set to 0
-    real_socket.settimeout.assert_called_once_with(0)
+    # And the socket was set to blocking
+    real_socket.setblocking.assert_called_once_with(1)
 
     # And recv was called with the bufsize
     real_socket.recv.assert_has_calls([
-        call(16),
-        call(16),
+        call(socket._bufsize)
     ])
 
     # And the buffer should contain the data from the server
-    socket.fd.getvalue().should.equal("response from server")
+    socket.fd.read().should.equal(b"response from server")
 
-    # And connect was never called
-    real_socket.connect.called.should.be.false
+    # And connect was called
+    real_socket.connect.called.should.be.true
 
 
 @patch('httpretty.core.old_socket')
 @patch('httpretty.core.socket')
-def test_fakesock_socket_real_sendall_continue_eagain(socket, old_socket):
+def test_fakesock_socket_real_sendall_continue_eagain_when_http(socket, old_socket):
     ("fakesock.socket#real_sendall should continue if the socket error was EAGAIN")
     socket.error = SocketErrorStub
     # Background: the real socket will stop returning bytes after the
     # first call
     real_socket = old_socket.return_value
-    real_socket.recv.side_effect = [SocketErrorStub(errno.EAGAIN), 'after error', ""]
+    real_socket.recv.side_effect = [SocketErrorStub(errno.EAGAIN), b'after error', b""]
 
     # Given a fake socket
     socket = fakesock.socket()
-
+    socket._address = ('1.2.3.4', 42)
+    socket.is_http = True
 
     # When I call real_sendall with data, some args and kwargs
-    socket.real_sendall("SOMEDATA", 'some extra args...', foo='bar')
+    socket.real_sendall(b"SOMEDATA", b'some extra args...', foo=b'bar')
 
     # Then it should have called sendall in the real socket
-    real_socket.sendall.assert_called_once_with("SOMEDATA", 'some extra args...', foo='bar')
+    real_socket.sendall.assert_called_once_with(b"SOMEDATA", b'some extra args...', foo=b'bar')
 
-    # And the timeout was set to 0
-    real_socket.settimeout.assert_called_once_with(0)
+    # And the socket was set to blocking
+    real_socket.setblocking.assert_called_once_with(1)
 
     # And recv was called with the bufsize
     real_socket.recv.assert_has_calls([
-        call(16),
-        call(16),
+        call(socket._bufsize)
     ])
 
     # And the buffer should contain the data from the server
-    socket.fd.getvalue().should.equal("after error")
+    socket.fd.read().should.equal(b"after error")
 
-    # And connect was never called
-    real_socket.connect.called.should.be.false
+    # And connect was called
+    real_socket.connect.called.should.be.true
 
 
 @patch('httpretty.core.old_socket')
 @patch('httpretty.core.socket')
-def test_fakesock_socket_real_sendall_socket_error(socket, old_socket):
+def test_fakesock_socket_real_sendall_socket_error_when_http(socket, old_socket):
     ("fakesock.socket#real_sendall should continue if the socket error was EAGAIN")
     socket.error = SocketErrorStub
     # Background: the real socket will stop returning bytes after the
     # first call
     real_socket = old_socket.return_value
-    real_socket.recv.side_effect = [SocketErrorStub(42), 'after error', ""]
+    real_socket.recv.side_effect = [SocketErrorStub(42), b'after error', ""]
 
     # Given a fake socket
     socket = fakesock.socket()
+    socket._address = ('1.2.3.4', 42)
+    socket.is_http = True
 
     # When I call real_sendall with data, some args and kwargs
-    socket.real_sendall("SOMEDATA", 'some extra args...', foo='bar')
+    socket.real_sendall(b"SOMEDATA", b'some extra args...', foo=b'bar')
 
     # Then it should have called sendall in the real socket
-    real_socket.sendall.assert_called_once_with("SOMEDATA", 'some extra args...', foo='bar')
+    real_socket.sendall.assert_called_once_with(b"SOMEDATA", b'some extra args...', foo=b'bar')
 
-    # And the timeout was set to 0
-    real_socket.settimeout.assert_called_once_with(0)
+    # And the socket was set to blocking
+    real_socket.setblocking.assert_called_once_with(1)
 
     # And recv was called with the bufsize
-    real_socket.recv.assert_called_once_with(16)
+    real_socket.recv.assert_called_once_with(socket._bufsize)
 
     # And the buffer should contain the data from the server
-    socket.fd.getvalue().should.equal("")
+    socket.fd.read().should.equal(b"")
 
-    # And connect was never called
-    real_socket.connect.called.should.be.false
+    # And connect was called
+    real_socket.connect.called.should.be.true
 
 
 @patch('httpretty.core.old_socket')
 @patch('httpretty.core.POTENTIAL_HTTP_PORTS')
-def test_fakesock_socket_real_sendall_when_http(POTENTIAL_HTTP_PORTS, old_socket):
+def test_fakesock_socket_real_sendall_when_sending_data(POTENTIAL_HTTP_PORTS, old_socket):
     ("fakesock.socket#real_sendall should connect before sending data")
     # Background: the real socket will stop returning bytes after the
     # first call
     real_socket = old_socket.return_value
-    real_socket.recv.side_effect = ['response from foobar :)', ""]
+    real_socket.recv.side_effect = [b'response from foobar :)', b""]
 
     # And the potential http port is 4000
     POTENTIAL_HTTP_PORTS.__contains__.side_effect = lambda other: int(other) == 4000
+    POTENTIAL_HTTP_PORTS.union.side_effect = lambda other: POTENTIAL_HTTP_PORTS
 
     # Given a fake socket
     socket = fakesock.socket()
@@ -415,22 +444,21 @@ def test_fakesock_socket_real_sendall_when_http(POTENTIAL_HTTP_PORTS, old_socket
     socket.connect(('foobar.com', 4000))
 
     # And send some data
-    socket.real_sendall("SOMEDATA")
+    socket.real_sendall(b"SOMEDATA")
 
     # Then connect should have been called
     real_socket.connect.assert_called_once_with(('foobar.com', 4000))
 
-    # And the timeout was set to 0
-    real_socket.settimeout.assert_called_once_with(0)
+    # And the socket was set to blocking
+    real_socket.setblocking.assert_called_once_with(1)
 
     # And recv was called with the bufsize
     real_socket.recv.assert_has_calls([
-        call(16),
-        call(16),
+        call(socket._bufsize)
     ])
 
     # And the buffer should contain the data from the server
-    socket.fd.getvalue().should.equal("response from foobar :)")
+    socket.fd.read().should.equal(b"response from foobar :)")
 
 
 @patch('httpretty.core.old_socket')
@@ -438,8 +466,8 @@ def test_fakesock_socket_real_sendall_when_http(POTENTIAL_HTTP_PORTS, old_socket
 @patch('httpretty.core.POTENTIAL_HTTP_PORTS')
 def test_fakesock_socket_sendall_with_valid_requestline(POTENTIAL_HTTP_PORTS, httpretty, old_socket):
     ("fakesock.socket#sendall should create an entry if it's given a valid request line")
-    matcher = Mock()
-    info = Mock()
+    matcher = Mock(name='matcher')
+    info = Mock(name='info')
     httpretty.match_uriinfo.return_value = (matcher, info)
     httpretty.register_uri(httpretty.GET, 'http://foo.com/foobar')
 
@@ -456,16 +484,16 @@ def test_fakesock_socket_sendall_with_valid_requestline(POTENTIAL_HTTP_PORTS, ht
     socket.connect(('foo.com', 80))
 
     # When I try to send data
-    socket.sendall("GET /foobar HTTP/1.1\r\nContent-Type: application/json\r\n\r\n")
+    socket.sendall(b"GET /foobar HTTP/1.1\r\nContent-Type: application/json\r\n\r\n")
 
 
 @patch('httpretty.core.old_socket')
 @patch('httpretty.core.httpretty')
 @patch('httpretty.core.POTENTIAL_HTTP_PORTS')
-def test_fakesock_socket_sendall_with_valid_requestline(POTENTIAL_HTTP_PORTS, httpretty, old_socket):
+def test_fakesock_socket_sendall_with_valid_requestline_2(POTENTIAL_HTTP_PORTS, httpretty, old_socket):
     ("fakesock.socket#sendall should create an entry if it's given a valid request line")
-    matcher = Mock()
-    info = Mock()
+    matcher = Mock(name='matcher')
+    info = Mock(name='info')
     httpretty.match_uriinfo.return_value = (matcher, info)
     httpretty.register_uri(httpretty.GET, 'http://foo.com/foobar')
 
@@ -482,18 +510,18 @@ def test_fakesock_socket_sendall_with_valid_requestline(POTENTIAL_HTTP_PORTS, ht
     socket.connect(('foo.com', 80))
 
     # When I try to send data
-    socket.sendall("GET /foobar HTTP/1.1\r\nContent-Type: application/json\r\n\r\n")
+    socket.sendall(b"GET /foobar HTTP/1.1\r\nContent-Type: application/json\r\n\r\n")
 
 
 @patch('httpretty.core.old_socket')
-@patch('httpretty.core.POTENTIAL_HTTP_PORTS')
-def test_fakesock_socket_sendall_with_body_data_no_entry(POTENTIAL_HTTP_PORTS, old_socket):
+def test_fakesock_socket_sendall_with_body_data_no_entry(old_socket):
     ("fakesock.socket#sendall should call real_sendall when not parsing headers and there is no entry")
     # Background:
     # Using a subclass of socket that mocks out real_sendall
+
     class MySocket(fakesock.socket):
         def real_sendall(self, data):
-            data.should.equal('BLABLABLABLA')
+            data.should.equal(b'BLABLABLABLA')
             return 'cool'
 
     # Given an instance of that socket
@@ -504,7 +532,7 @@ def test_fakesock_socket_sendall_with_body_data_no_entry(POTENTIAL_HTTP_PORTS, o
     socket.connect(('foo.com', 80))
 
     # When I try to send data
-    result = socket.sendall("BLABLABLABLA")
+    result = socket.sendall(b"BLABLABLABLA")
 
     # Then the result should be the return value from real_sendall
     result.should.equal('cool')
@@ -513,47 +541,53 @@ def test_fakesock_socket_sendall_with_body_data_no_entry(POTENTIAL_HTTP_PORTS, o
 @patch('httpretty.core.old_socket')
 @patch('httpretty.core.POTENTIAL_HTTP_PORTS')
 def test_fakesock_socket_sendall_with_body_data_with_entry(POTENTIAL_HTTP_PORTS, old_socket):
-    ("fakesock.socket#sendall should call real_sendall when not ")
+    ("fakesock.socket#sendall should call real_sendall when there is no entry")
     # Background:
     # Using a subclass of socket that mocks out real_sendall
+    data_sent = []
+
     class MySocket(fakesock.socket):
         def real_sendall(self, data):
-            raise AssertionError('should have never been called')
-    # Using a mocked entry
-    entry = Mock()
-    entry.request.headers = {}
-    entry.request.body = ''
+            data_sent.append(data)
 
     # Given an instance of that socket
     socket = MySocket()
-    socket._entry = entry
-
 
     # And that is is considered http
     socket.connect(('foo.com', 80))
 
     # When I try to send data
-    socket.sendall("BLABLABLABLA")
+    socket.sendall(b"BLABLABLABLA")
 
-    # Then the entry should have that body
-    entry.request.body.should.equal('BLABLABLABLA')
+    # Then it should have called real_sendall
+    data_sent.should.equal([b'BLABLABLABLA'])
 
 
+@patch('httpretty.core.httpretty.match_uriinfo')
 @patch('httpretty.core.old_socket')
 @patch('httpretty.core.POTENTIAL_HTTP_PORTS')
-def test_fakesock_socket_sendall_with_body_data_with_chunked_entry(POTENTIAL_HTTP_PORTS, old_socket):
+def test_fakesock_socket_sendall_with_body_data_with_chunked_entry(POTENTIAL_HTTP_PORTS, old_socket, match_uriinfo):
     ("fakesock.socket#sendall should call real_sendall when not ")
     # Background:
     # Using a subclass of socket that mocks out real_sendall
+
     class MySocket(fakesock.socket):
         def real_sendall(self, data):
             raise AssertionError('should have never been called')
+
+    matcher = Mock(name='matcher')
+    info = Mock(name='info')
+    httpretty.match_uriinfo.return_value = (matcher, info)
+
     # Using a mocked entry
     entry = Mock()
+    entry.method = 'GET'
+    entry.info.path = '/foo'
+
     entry.request.headers = {
         'transfer-encoding': 'chunked',
     }
-    entry.request.body = ''
+    entry.request.body = b''
 
     # Given an instance of that socket
     socket = MySocket()
@@ -563,7 +597,72 @@ def test_fakesock_socket_sendall_with_body_data_with_chunked_entry(POTENTIAL_HTT
     socket.connect(('foo.com', 80))
 
     # When I try to send data
-    socket.sendall("BLABLABLABLA")
+    socket.sendall(b"BLABLABLABLA")
 
     # Then the entry should have that body
-    httpretty.last_request.body.should.equal('BLABLABLABLA')
+    httpretty.last_request.body.should.equal(b'BLABLABLABLA')
+
+
+def test_fakesock_socket_sendall_with_path_starting_with_two_slashes():
+    ("fakesock.socket#sendall handles paths starting with // well")
+
+    httpretty.register_uri(httpretty.GET, 'http://example.com//foo')
+
+    class MySocket(fakesock.socket):
+        def real_sendall(self, data, *args, **kw):
+            raise AssertionError('should never call this...')
+
+    # Given an instance of that socket
+    socket = MySocket()
+
+    # And that is is considered http
+    socket.connect(('example.com', 80))
+
+    # When I try to send data
+    socket.sendall(b"GET //foo HTTP/1.1\r\nContent-Type: application/json\r\n\r\n")
+
+
+def test_URIMatcher_respects_querystring():
+    ("URIMatcher response querystring")
+    matcher = URIMatcher('http://www.foo.com/?query=true', None)
+    info = URIInfo.from_uri('http://www.foo.com/', None)
+    assert matcher.matches(info)
+
+    matcher = URIMatcher('http://www.foo.com/?query=true', None, match_querystring=True)
+    info = URIInfo.from_uri('http://www.foo.com/', None)
+    assert not matcher.matches(info)
+
+    matcher = URIMatcher('http://www.foo.com/?query=true', None, match_querystring=True)
+    info = URIInfo.from_uri('http://www.foo.com/?query=true', None)
+    assert matcher.matches(info)
+
+    matcher = URIMatcher('http://www.foo.com/?query=true&unquery=false', None, match_querystring=True)
+    info = URIInfo.from_uri('http://www.foo.com/?unquery=false&query=true', None)
+    assert matcher.matches(info)
+
+    matcher = URIMatcher('http://www.foo.com/?unquery=false&query=true', None, match_querystring=True)
+    info = URIInfo.from_uri('http://www.foo.com/?query=true&unquery=false', None)
+    assert matcher.matches(info)
+
+
+def test_URIMatcher_equality_respects_querystring():
+    ("URIMatcher equality check should check querystring")
+    matcher_a = URIMatcher('http://www.foo.com/?query=true', None)
+    matcher_b = URIMatcher('http://www.foo.com/?query=false', None)
+    assert matcher_a == matcher_b
+
+    matcher_a = URIMatcher('http://www.foo.com/?query=true', None)
+    matcher_b = URIMatcher('http://www.foo.com/', None)
+    assert matcher_a == matcher_b
+
+    matcher_a = URIMatcher('http://www.foo.com/?query=true', None, match_querystring=True)
+    matcher_b = URIMatcher('http://www.foo.com/?query=false', None, match_querystring=True)
+    assert not matcher_a == matcher_b
+
+    matcher_a = URIMatcher('http://www.foo.com/?query=true', None, match_querystring=True)
+    matcher_b = URIMatcher('http://www.foo.com/', None, match_querystring=True)
+    assert not matcher_a == matcher_b
+
+    matcher_a = URIMatcher('http://www.foo.com/?query=true&unquery=false', None, match_querystring=True)
+    matcher_b = URIMatcher('http://www.foo.com/?unquery=false&query=true', None, match_querystring=True)
+    assert matcher_a == matcher_b
